from typing import Tuple
import numpy as np

from pyquil import Program, gates
from pyquil.api import WavefunctionSimulator

from .gates_pyquil import PyquilGateApplicator
from openqaoa.backends.basebackend import QAOABaseBackendStatevector
from openqaoa.qaoa_components import QAOADescriptor
from openqaoa.qaoa_components.variational_parameters.variational_baseparams import (
    QAOAVariationalBaseParams,
)
from openqaoa.qaoa_components.ansatz_constructor.gatemap import (
    RXGateMap,
    RYGateMap,
    RZGateMap,
)
from openqaoa.backends.cost_function import cost_function
from openqaoa.utilities import generate_uuid, round_value


class QAOAPyQuilWavefunctionSimulatorBackend(QAOABaseBackendStatevector):
    """
    A local Wavefunction simulator backend for the PyQuil service provider
    """

    PYQUIL_ROTATIONGATES_LIBRARY = [RXGateMap, RYGateMap, RZGateMap]

    def __init__(
        self,
        qaoa_descriptor: QAOADescriptor,
        prepend_state: Program = None,
        append_state: Program = None,
        init_hadamard: bool = True,
        cvar_alpha: float = 1,
    ):
        QAOABaseBackendStatevector.__init__(
            self,
            qaoa_descriptor,
            prepend_state,
            append_state,
            init_hadamard,
            cvar_alpha,
        )

    def qaoa_circuit(self, params: QAOAVariationalBaseParams) -> Program:
        """
        Creates a QAOA circuit (pyquil.Program object), given the qubit pairs, single qubits with biases,
        and a set of circuit angles. Note that this function does not actually run
        the circuit. To do this, you will need to subsequently execute the command self.eng.flush().

        Parameters
        ----------
        params: `QAOAVariationalBaseParams`

        Returns
        -------
        `pyquil.Program`
                A pyquil.Program object.
        """
        gates_applicator = PyquilGateApplicator()

        self.assign_angles(params)

        circuit = Program()
        if self.prepend_state:
            circuit += self.prepend_state

        # Initial state is all |+>
        if self.init_hadamard:
            for i in range(self.n_qubits):
                circuit += gates.H(i)

        # create a list of gates in order of application on quantum circuit
        low_level_gate_list = []
        for i, each_gate in enumerate(self.abstract_circuit):
            if (
                type(each_gate)
                in QAOAPyQuilWavefunctionSimulatorBackend.PYQUIL_ROTATIONGATES_LIBRARY
            ):
                decomposition = each_gate.decomposition("trivial")
            else:
                decomposition = each_gate.decomposition("standard")
            # using the list above, construct the circuit
            for each_tuple in decomposition:
                gate = each_tuple[0](gates_applicator, *each_tuple[1])
                gate.apply_gate(circuit)

        if self.append_state:
            circuit += self.append_state

        return circuit

    def wavefunction(self, params: QAOAVariationalBaseParams):
        """
        Get the wavefunction of the state produced by the parametric circuit.

        Parameters
        ----------
        params: `QAOAVariationalBaseParams`
                The QAOA parameters - an object of one of the parameter classes, containing
                the variational parameters (angles).

        Returns
        -------
        wf: `List[complex]`
                pyquil Wavefunction object.
        """
        # generate a job id for the wavefunction evaluation
        self.job_id = generate_uuid()

        program = self.qaoa_circuit(params)

        wf_sim = WavefunctionSimulator()
        wf = wf_sim.wavefunction(program)
        self.measurement_outcomes = wf.amplitudes
        return wf

    @round_value
    def expectation(self, params: QAOAVariationalBaseParams) -> float:
        """
        Compute the expectation value w.r.t the Cost Hamiltonian

        Parameters
        ----------
        params: `QAOAVariationalBaseParams`
                The QAOA parameters - an object of one of the parameter classes, containing
                variable parameters.

        Returns
        -------
        `float`
                expectation value of cost operator wrt to quantum state produced by QAOA circuit
        """
        prob_dict = self.probability_dict(params)
        cost = cost_function(
            prob_dict, self.qaoa_descriptor.cost_hamiltonian, self.cvar_alpha
        )
        return cost

    @round_value
    def expectation_w_uncertainty(
        self, params: QAOAVariationalBaseParams
    ) -> Tuple[float, float]:
        """
        Compute the expectation value w.r.t the Cost Hamiltonian and its uncertainty

        Parameters
        ----------
        params: `QAOAVariationalBaseParams`
                The QAOA parameters - an object of one of the parameter classes, containing
                variable parameters.

        Returns
        -------
        `Tuple[float]`
                expectation value and its uncertainty of cost operator wrt
                to quantum state produced by QAOA circuit.
        """
        prob_dict = self.probability_dict(params)
        cost = cost_function(
            prob_dict, self.qaoa_descriptor.cost_hamiltonian, self.cvar_alpha
        )
        cost_sq = cost_function(
            prob_dict,
            self.qaoa_descriptor.cost_hamiltonian.hamiltonian_squared,
            self.cvar_alpha,
        )

        uncertainty = np.sqrt(cost_sq - cost**2)

        return (cost, uncertainty)

    def circuit_to_qasm(self):
        raise NotImplementedError()

    def reset_circuit(self):
        raise NotImplementedError()



# unit testing for circuit routing functionality in OQ
import unittest
import numpy as np
from typing import List, Optional
import pytest

from openqaoa import QAOA, create_device, QUBO
from openqaoa.qaoa_components import (
    create_qaoa_variational_params,
    QAOADescriptor,
    PauliOp,
    Hamiltonian,
)
from openqaoa.utilities import X_mixer_hamiltonian
from openqaoa.backends import create_device
from openqaoa.problems import MaximumCut, ShortestPath
from openqaoa_pyquil.backends import DevicePyquil, QAOAPyQuilQPUBackend
from openqaoa.backends.devices_core import DeviceBase
from openqaoa.qaoa_components.ansatz_constructor.gatemap import SWAPGateMap


class TestingQAOAPyquilQVM_QR(unittest.TestCase):

    """Tests pyquil backend compatibility with routing_function.

    For all of these tests, qvm and quilc must be running.
    """

    def test_no_swap(self):
        """
        Tests that QAOADescriptor with a trivial `routing_function` input (with no swaps) returns identical
        results as QAOADescriptor with no `routing_function` input, by comparing output of seeded QVM run.
        Different values of p, arguments, and cost hamiltonian coefficients are tested.

        """

        def routing_function_test1(device, problem_to_solve):
            # tuples ordered from 0,n, both SWAP and ising gates
            gate_list_indices = [[0, 1]]

            # True for SWAP
            swap_mask = [False]

            # {QPU: (0 to n index)}
            initial_physical_to_logical_mapping = {0: 0, 1: 1}

            # 0 to n, permuted
            final_mapping = [0, 1]

            return (
                gate_list_indices,
                swap_mask,
                initial_physical_to_logical_mapping,
                final_mapping,
            )

        args_lst = [
            [np.pi / 8, np.pi / 4],
            [np.pi / 3.5, np.pi / 3],
            [np.pi / 8, np.pi / 4],
            [np.pi / 3.5, np.pi / 3],
            [1, 2, 3, 4],
            [np.pi / 8, np.pi / 4, np.pi / 8, np.pi / 4],
            [1, 2, 3, 4],
            [np.pi / 8, np.pi / 4, np.pi / 8, np.pi / 4],
        ]
        p_lst = [1, 1, 1, 1, 2, 2, 2, 2]
        cost_hamil_lst = [
            Hamiltonian(
                [PauliOp("Z", (0,)), PauliOp("Z", (1,)), PauliOp("ZZ", (0, 1))],
                [1, 1, 1],
                1,
            ),
            Hamiltonian(
                [PauliOp("Z", (0,)), PauliOp("Z", (1,)), PauliOp("ZZ", (0, 1))],
                [1, 1, 1],
                1,
            ),
            Hamiltonian(
                [PauliOp("Z", (0,)), PauliOp("Z", (1,)), PauliOp("ZZ", (0, 1))],
                [0.5, 0, 2],
                0.7,
            ),
            Hamiltonian(
                [PauliOp("Z", (0,)), PauliOp("Z", (1,)), PauliOp("ZZ", (0, 1))],
                [1, 1.2, 1],
                0,
            ),
            Hamiltonian(
                [PauliOp("Z", (0,)), PauliOp("Z", (1,)), PauliOp("ZZ", (0, 1))],
                [1, 1, 1],
                1,
            ),
            Hamiltonian(
                [PauliOp("Z", (0,)), PauliOp("Z", (1,)), PauliOp("ZZ", (0, 1))],
                [1, 1, 1],
                1,
            ),
            Hamiltonian(
                [PauliOp("Z", (0,)), PauliOp("Z", (1,)), PauliOp("ZZ", (0, 1))],
                [0.5, 0, 2],
                0.7,
            ),
            Hamiltonian(
                [PauliOp("Z", (0,)), PauliOp("Z", (1,)), PauliOp("ZZ", (0, 1))],
                [1, 1.2, 1],
                0,
            ),
        ]
        shots = 2
        seed = 1

        device_pyquil = DevicePyquil(
            device_name="2q-qvm", as_qvm=True, execution_timeout=5, compiler_timeout=5
        )
        device_pyquil.quantum_computer.qam.random_seed = seed

        for i in range(len(p_lst)):
            p = p_lst[i]
            args = args_lst[i]
            cost_hamil = cost_hamil_lst[i]

            # With routing
            mixer_hamil = X_mixer_hamiltonian(n_qubits=2)
            qaoa_descriptor = QAOADescriptor(
                cost_hamil, mixer_hamil, routing_function=routing_function_test1, p=p
            )
            variate_params = create_qaoa_variational_params(
                qaoa_descriptor, "standard", "ramp"
            )

            variate_params.update_from_raw(args)
            backend_obj_pyquil = QAOAPyQuilQPUBackend(
                qaoa_descriptor=qaoa_descriptor,
                device=device_pyquil,
                prepend_state=None,
                append_state=None,
                init_hadamard=True,
                cvar_alpha=1,
                n_shots=shots,
            )
            expt_pyquil_w_qr = backend_obj_pyquil.expectation(variate_params)

            # No routing
            mixer_hamil = X_mixer_hamiltonian(n_qubits=2)
            qaoa_descriptor = QAOADescriptor(cost_hamil, mixer_hamil, p=p)
            variate_params = create_qaoa_variational_params(
                qaoa_descriptor, "standard", "ramp"
            )

            variate_params.update_from_raw(args)
            backend_obj_pyquil = QAOAPyQuilQPUBackend(
                qaoa_descriptor=qaoa_descriptor,
                device=device_pyquil,
                prepend_state=None,
                append_state=None,
                init_hadamard=True,
                cvar_alpha=1,
                n_shots=shots,
            )
            expt_pyquil_no_qr = backend_obj_pyquil.expectation(variate_params)

            self.assertAlmostEqual(expt_pyquil_w_qr, expt_pyquil_no_qr)

    def test_cancelled_swap(self):
        """
        Tests that QAOADescriptor with a trivial `routing_function` input (with two swaps that cancel each other)
        returns identical results as QAOADescriptor with no `routing_function` input,
        by comparing output of seeded QVM run. Different values of p, arguments,
        and cost hamiltonian coefficients are tested.
        """

        def routing_function_test1(device, problem_to_solve):
            # tuples ordered from 0,n, both SWAP and ising gates
            gate_list_indices = [[0, 1], [0, 1], [0, 1]]

            # True for SWAP
            swap_mask = [True, True, False]

            # {QPU: (0 to n index)}
            initial_physical_to_logical_mapping = {0: 0, 1: 1}

            # 0 to n, permuted
            final_mapping = [0, 1]

            return (
                gate_list_indices,
                swap_mask,
                initial_physical_to_logical_mapping,
                final_mapping,
            )

        args_lst = [
            [np.pi / 8, np.pi / 4],
            [np.pi / 3.5, np.pi / 3],
            [np.pi / 8, np.pi / 4],
            [np.pi / 3.5, np.pi / 3],
            [1, 2, 3, 4],
            [np.pi / 8, np.pi / 4, np.pi / 8, np.pi / 4],
            [1, 2, 3, 4],
            [np.pi / 8, np.pi / 4, np.pi / 8, np.pi / 4],
        ]
        p_lst = [1, 1, 1, 1, 2, 2, 2, 2]
        cost_hamil_lst = [
            Hamiltonian(
                [PauliOp("Z", (0,)), PauliOp("Z", (1,)), PauliOp("ZZ", (0, 1))],
                [1, 1, 1],
                1,
            ),
            Hamiltonian(
                [PauliOp("Z", (0,)), PauliOp("Z", (1,)), PauliOp("ZZ", (0, 1))],
                [1, 1, 1],
                1,
            ),
            Hamiltonian(
                [PauliOp("Z", (0,)), PauliOp("Z", (1,)), PauliOp("ZZ", (0, 1))],
                [0.5, 0, 2],
                0.7,
            ),
            Hamiltonian(
                [PauliOp("Z", (0,)), PauliOp("Z", (1,)), PauliOp("ZZ", (0, 1))],
                [1.5, 1.2, 1],
                0,
            ),
            Hamiltonian(
                [PauliOp("Z", (0,)), PauliOp("Z", (1,)), PauliOp("ZZ", (0, 1))],
                [1, 1, 1],
                1,
            ),
            Hamiltonian(
                [PauliOp("Z", (0,)), PauliOp("Z", (1,)), PauliOp("ZZ", (0, 1))],
                [1, 1, 1],
                1,
            ),
            Hamiltonian(
                [PauliOp("Z", (0,)), PauliOp("Z", (1,)), PauliOp("ZZ", (0, 1))],
                [0.5, 0, 2],
                0.7,
            ),
            Hamiltonian(
                [PauliOp("Z", (0,)), PauliOp("Z", (1,)), PauliOp("ZZ", (0, 1))],
                [1, 5.2, 1],
                0,
            ),
        ]
        shots = 3
        seed = 4

        device_pyquil = DevicePyquil(
            device_name="2q-qvm", as_qvm=True, execution_timeout=5, compiler_timeout=5
        )
        device_pyquil.quantum_computer.qam.random_seed = seed

        for i in range(len(p_lst)):
            p = p_lst[i]
            args = args_lst[i]
            cost_hamil = cost_hamil_lst[i]

            # With routing
            mixer_hamil = X_mixer_hamiltonian(n_qubits=2)
            qaoa_descriptor = QAOADescriptor(
                cost_hamil, mixer_hamil, routing_function=routing_function_test1, p=p
            )
            variate_params = create_qaoa_variational_params(
                qaoa_descriptor, "standard", "ramp"
            )

            variate_params.update_from_raw(args)
            backend_obj_pyquil = QAOAPyQuilQPUBackend(
                qaoa_descriptor=qaoa_descriptor,
                device=device_pyquil,
                prepend_state=None,
                append_state=None,
                init_hadamard=True,
                cvar_alpha=1,
                n_shots=shots,
            )
            expt_pyquil_w_qr = backend_obj_pyquil.expectation(variate_params)

            # No routing
            mixer_hamil = X_mixer_hamiltonian(n_qubits=2)
            qaoa_descriptor = QAOADescriptor(cost_hamil, mixer_hamil, p=p)
            variate_params = create_qaoa_variational_params(
                qaoa_descriptor, "standard", "ramp"
            )

            variate_params.update_from_raw(args)
            backend_obj_pyquil = QAOAPyQuilQPUBackend(
                qaoa_descriptor=qaoa_descriptor,
                device=device_pyquil,
                prepend_state=None,
                append_state=None,
                init_hadamard=True,
                cvar_alpha=1,
                n_shots=shots,
            )
            expt_pyquil_no_qr = backend_obj_pyquil.expectation(variate_params)

            self.assertAlmostEqual(expt_pyquil_w_qr, expt_pyquil_no_qr)

    def test_simplest_swap(self):
        """
        Tests that QAOADescriptor with a trivial `routing_function` input (with no swaps) returns identical
        results as QAOADescriptor with no `routing_function` input, by comparing output of seeded QVM run.
        Different values of p, arguments, and cost hamiltonian coefficients are tested.

        Note : Even with a fixed seed, insertion of swaps changes measurement statistics.
        Final assertion is therefore only up to a tolerance, chosen by eyeballing results for a chosen seed.

        """

        def routing_function_test1(device, problem_to_solve):
            # tuples ordered from 0,n, both SWAP and ising gates
            gate_list_indices = [[0, 1], [0, 1]]

            # True for SWAP
            swap_mask = [True, False]

            # {QPU: (0 to n index)}
            initial_physical_to_logical_mapping = {0: 0, 1: 1}

            # 0 to n, permuted
            final_mapping = [1, 0]

            return (
                gate_list_indices,
                swap_mask,
                initial_physical_to_logical_mapping,
                final_mapping,
            )

        args_lst = [
            [np.pi / 8, np.pi / 4],
            [np.pi / 3.5, np.pi / 3],
            [np.pi / 8, np.pi / 4],
            [np.pi / 3.5, np.pi / 3],
            [1, 2, 3, 4],
            [np.pi / 8, np.pi / 4, np.pi / 8, np.pi / 4],
            [1, 2, 3, 4],
            [np.pi / 8, np.pi / 4, np.pi / 8, np.pi / 4],
        ]
        p_lst = [1, 1, 1, 1, 2, 2, 2, 2]
        cost_hamil_lst = [
            Hamiltonian(
                [PauliOp("Z", (0,)), PauliOp("Z", (1,)), PauliOp("ZZ", (0, 1))],
                [1, 1, 1],
                1,
            ),
            Hamiltonian(
                [PauliOp("Z", (0,)), PauliOp("Z", (1,)), PauliOp("ZZ", (0, 1))],
                [1, 1, 1],
                1,
            ),
            Hamiltonian(
                [PauliOp("Z", (0,)), PauliOp("Z", (1,)), PauliOp("ZZ", (0, 1))],
                [0.5, 0, 2],
                0.7,
            ),
            Hamiltonian(
                [PauliOp("Z", (0,)), PauliOp("Z", (1,)), PauliOp("ZZ", (0, 1))],
                [1.5, 1.2, 1],
                0,
            ),
            Hamiltonian(
                [PauliOp("Z", (0,)), PauliOp("Z", (1,)), PauliOp("ZZ", (0, 1))],
                [1, 1, 1],
                1,
            ),
            Hamiltonian(
                [PauliOp("Z", (0,)), PauliOp("Z", (1,)), PauliOp("ZZ", (0, 1))],
                [1, 1, 1],
                1,
            ),
            Hamiltonian(
                [PauliOp("Z", (0,)), PauliOp("Z", (1,)), PauliOp("ZZ", (0, 1))],
                [0.5, 0, 2],
                0.7,
            ),
            Hamiltonian(
                [PauliOp("Z", (0,)), PauliOp("Z", (1,)), PauliOp("ZZ", (0, 1))],
                [1, 5.2, 1],
                0,
            ),
        ]
        shots = 10
        seed = 4

        device_pyquil = DevicePyquil(
            device_name="2q-qvm", as_qvm=True, execution_timeout=10, compiler_timeout=10
        )
        device_pyquil.quantum_computer.qam.random_seed = seed

        for i in range(len(p_lst)):
            p = p_lst[i]
            args = args_lst[i]
            cost_hamil = cost_hamil_lst[i]

            # With routing
            mixer_hamil = X_mixer_hamiltonian(n_qubits=2)
            qaoa_descriptor = QAOADescriptor(
                cost_hamil, mixer_hamil, routing_function=routing_function_test1, p=p
            )
            variate_params = create_qaoa_variational_params(
                qaoa_descriptor, "standard", "ramp"
            )

            variate_params.update_from_raw(args)
            backend_obj_pyquil = QAOAPyQuilQPUBackend(
                qaoa_descriptor=qaoa_descriptor,
                device=device_pyquil,
                prepend_state=None,
                append_state=None,
                init_hadamard=True,
                cvar_alpha=1,
                n_shots=shots,
            )
            expt_pyquil_w_qr = backend_obj_pyquil.expectation(variate_params)

            # No routing
            mixer_hamil = X_mixer_hamiltonian(n_qubits=2)
            qaoa_descriptor = QAOADescriptor(cost_hamil, mixer_hamil, p=p)
            variate_params = create_qaoa_variational_params(
                qaoa_descriptor, "standard", "ramp"
            )

            variate_params.update_from_raw(args)
            backend_obj_pyquil = QAOAPyQuilQPUBackend(
                qaoa_descriptor=qaoa_descriptor,
                device=device_pyquil,
                prepend_state=None,
                append_state=None,
                init_hadamard=True,
                cvar_alpha=1,
                n_shots=shots,
            )
            expt_pyquil_no_qr = backend_obj_pyquil.expectation(variate_params)

            # Note : Even with a fixed seed, insertion of swaps changes measurement statistics.
            # Final assertion is therefore only up to a tolerance, chosen by eyeballing results for a chosen seed.
            self.assertAlmostEqual(expt_pyquil_w_qr, expt_pyquil_no_qr, delta=1)

    def test_different_topologies(self):
        """
        Tests QAOADescriptor with different devices.
        results as QAOADescriptor with no `routing_function` input, by comparing output of seeded QVM run.
        Different values of p, arguments, and cost hamiltonian coefficients are tested.

        """

        def routing_function_test1(device, problem_to_solve):
            # tuples ordered from 0,n, both SWAP and ising gates
            gate_list_indices = [[0, 1], [1, 0], [0, 1]]

            # True for SWAP
            swap_mask = [True, True, False]

            # {QPU: (0 to n index)}
            initial_physical_to_logical_mapping = {0: 0, 1: 1}

            # 0 to n, permuted
            final_mapping = [0, 1]

            return (
                gate_list_indices,
                swap_mask,
                initial_physical_to_logical_mapping,
                final_mapping,
            )

        args_lst = [[np.pi / 8, np.pi / 4], [np.pi / 8, np.pi / 4, 1, 2], [1, 2, 3, 4]]
        p_lst = [1, 2, 2]
        cost_hamil_lst = [
            Hamiltonian(
                [PauliOp("Z", (0,)), PauliOp("Z", (1,)), PauliOp("ZZ", (0, 1))],
                [1, 1.5, 2],
                0.5,
            ),
            Hamiltonian(
                [PauliOp("Z", (0,)), PauliOp("Z", (1,)), PauliOp("ZZ", (0, 1))],
                [1, 1.5, 2],
                0.5,
            ),
            Hamiltonian(
                [PauliOp("Z", (0,)), PauliOp("Z", (1,)), PauliOp("ZZ", (0, 1))],
                [1, 1.5, 2],
                0.5,
            ),
        ]

        device_name_lst = ["2q-qvm", "3q-qvm", "Aspen-M-3"]

        shots = 2
        seed = 1

        for i in range(len(p_lst)):
            p = p_lst[i]
            args = args_lst[i]
            cost_hamil = cost_hamil_lst[i]

            device_pyquil = DevicePyquil(
                device_name=device_name_lst[i],
                as_qvm=True,
                execution_timeout=5,
                compiler_timeout=5,
            )
            device_pyquil.quantum_computer.qam.random_seed = seed

            # With routing
            mixer_hamil = X_mixer_hamiltonian(n_qubits=2)
            qaoa_descriptor = QAOADescriptor(
                cost_hamil, mixer_hamil, routing_function=routing_function_test1, p=p
            )
            variate_params = create_qaoa_variational_params(
                qaoa_descriptor, "standard", "ramp"
            )

            variate_params.update_from_raw(args)
            backend_obj_pyquil = QAOAPyQuilQPUBackend(
                qaoa_descriptor=qaoa_descriptor,
                device=device_pyquil,
                prepend_state=None,
                append_state=None,
                init_hadamard=True,
                cvar_alpha=1,
                n_shots=shots,
            )
            expt_pyquil_w_qr = backend_obj_pyquil.expectation(variate_params)

            # No routing
            mixer_hamil = X_mixer_hamiltonian(n_qubits=2)
            qaoa_descriptor = QAOADescriptor(cost_hamil, mixer_hamil, p=p)
            variate_params = create_qaoa_variational_params(
                qaoa_descriptor, "standard", "ramp"
            )

            variate_params.update_from_raw(args)
            backend_obj_pyquil = QAOAPyQuilQPUBackend(
                qaoa_descriptor=qaoa_descriptor,
                device=device_pyquil,
                prepend_state=None,
                append_state=None,
                init_hadamard=True,
                cvar_alpha=1,
                n_shots=shots,
            )
            expt_pyquil_no_qr = backend_obj_pyquil.expectation(variate_params)

            self.assertAlmostEqual(expt_pyquil_w_qr, expt_pyquil_no_qr)


class ExpectedRouting:
    def __init__(
        self,
        qubo,
        device_name,
        device_location,
        qpu_credentials,
        problem_to_solve,
        initial_mapping,
        gate_indices_list,
        swap_mask,
        initial_physical_to_logical_mapping,
        final_logical_qubit_order,
    ):
        self.qubo = qubo
        self.device = create_device(
            name=device_name, location=device_location, **qpu_credentials
        )

        self.device_name = device_name
        self.device_location = device_location
        self.problem_to_solve = problem_to_solve
        self.initial_mapping = initial_mapping

        self.gate_indices_list = gate_indices_list
        self.swap_mask = swap_mask
        self.initial_physical_to_logical_mapping = initial_physical_to_logical_mapping
        self.final_logical_qubit_order = final_logical_qubit_order

    def values_input(self):
        return (
            self.device_name,
            self.device_location,
            self.problem_to_solve,
            self.initial_mapping,
        )

    def values_return(self):
        return (
            self.gate_indices_list,
            self.swap_mask,
            self.initial_physical_to_logical_mapping,
            self.final_logical_qubit_order,
        )


class TestingQubitRouting(unittest.TestCase):
    def setUp(self):
        # case qubits device > qubits problem (RIGETTI)
        self.RIGETTI_SHORTESTPATH = ExpectedRouting(
            qubo=ShortestPath.random_instance(
                n_nodes=4, edge_probability=0.9, seed=20
            ).qubo,
            device_location="qcs",
            device_name="9q-square-qvm",
            qpu_credentials={
                "as_qvm": True,
                "execution_timeout": 10,
                "compiler_timeout": 100,
            },
            problem_to_solve=[
                (2, 3),
                (0, 2),
                (2, 4),
                (2, 5),
                (0, 4),
                (4, 5),
                (0, 5),
                (1, 3),
                (3, 5),
                (1, 5),
            ],
            initial_mapping=None,
            gate_indices_list=[
                [4, 5],
                [2, 4],
                [1, 3],
                [3, 5],
                [4, 5],
                [1, 4],
                [2, 4],
                [0, 6],
                [2, 6],
                [2, 6],
                [2, 4],
                [4, 5],
                [2, 4],
                [1, 3],
                [1, 6],
            ],
            swap_mask=[
                False,
                False,
                False,
                False,
                True,
                False,
                False,
                True,
                False,
                True,
                False,
                True,
                False,
                True,
                False,
            ],
            initial_physical_to_logical_mapping={
                8: 0,
                4: 1,
                6: 2,
                1: 3,
                3: 4,
                0: 5,
                7: 6,
            },
            final_logical_qubit_order=[2, 3, 6, 1, 4, 5, 0],
        )

        # case qubits device == qubits problem (RIGETTI)
        self.RIGETTI_MAXCUT = ExpectedRouting(
            qubo=MaximumCut.random_instance(
                n_nodes=9, edge_probability=0.9, seed=20
            ).qubo,
            device_location="qcs",
            device_name="9q-square-qvm",
            qpu_credentials={
                "as_qvm": True,
                "execution_timeout": 10,
                "compiler_timeout": 100,
            },
            problem_to_solve=[
                (0, 2),
                (0, 3),
                (0, 4),
                (0, 5),
                (0, 7),
                (0, 8),
                (1, 2),
                (1, 4),
                (1, 5),
                (1, 6),
                (1, 7),
                (1, 8),
                (2, 3),
                (2, 5),
                (2, 6),
                (2, 7),
                (2, 8),
                (3, 4),
                (3, 5),
                (3, 6),
                (3, 8),
                (4, 7),
                (4, 8),
                (5, 6),
                (5, 7),
                (5, 8),
                (6, 7),
                (6, 8),
                (7, 8),
            ],
            initial_mapping=None,
            gate_indices_list=[
                [5, 6],
                [3, 5],
                [0, 3],
                [1, 7],
                [0, 4],
                [1, 8],
                [4, 7],
                [6, 8],
                [6, 7],
                [1, 2],
                [2, 4],
                [4, 7],
                [0, 4],
                [1, 2],
                [6, 7],
                [5, 6],
                [1, 7],
                [3, 7],
                [4, 7],
                [6, 8],
                [0, 3],
                [0, 4],
                [3, 5],
                [5, 6],
                [6, 8],
                [3, 5],
                [1, 8],
                [1, 2],
                [6, 8],
                [1, 7],
                [3, 7],
                [4, 7],
                [4, 7],
                [6, 7],
                [0, 4],
                [0, 4],
                [2, 4],
            ],
            swap_mask=[
                False,
                False,
                False,
                False,
                False,
                False,
                False,
                False,
                False,
                False,
                True,
                False,
                False,
                False,
                True,
                False,
                False,
                False,
                False,
                False,
                True,
                False,
                False,
                True,
                False,
                False,
                True,
                False,
                False,
                True,
                False,
                False,
                True,
                False,
                False,
                True,
                False,
            ],
            initial_physical_to_logical_mapping={
                2: 0,
                7: 1,
                8: 2,
                1: 3,
                5: 4,
                0: 5,
                3: 6,
                4: 7,
                6: 8,
            },
            final_logical_qubit_order=[3, 8, 7, 4, 2, 6, 1, 5, 0],
        )

        # create a list of all the cases
        self.list_of_cases = []
        for value in self.__dict__.values():
            if isinstance(value, ExpectedRouting):
                self.list_of_cases.append(value)

    def __routing_function_mock(
        self,
        device: DeviceBase,
        problem_to_solve: List[List[int]],
        initial_mapping: Optional[List[int]] = None,
    ):
        """
        function that imitates the routing function for testing purposes.
        """

        for case in self.list_of_cases:
            if case.values_input() == (
                device.device_name,
                device.device_location,
                problem_to_solve,
                initial_mapping,
            ):
                return case.values_return()

        raise ValueError(
            """The input values are not in the list of expected values, 
                            check the expected cases and the input values. 
                            The input values are: device_name: {}, device_location: {}, problem_to_solve: {}, 
                            initial_mapping: {}""".format(
                device.device_name,
                device.device_location,
                problem_to_solve,
                initial_mapping,
            )
        )

    def __compare_results(self, expected: ExpectedRouting, p: int):
        """
        function that runs qaoa with the routing function and the problem to solve and
        compares the expected and actual results of the routing function.

        :param expected: ExpectedRouting object that contains the input and the expected results of the routing function
        :param p: number of layers of the qaoa circuit
        """
        print(f"\t Testing p={p}")

        device = expected.device
        qubo = expected.qubo

        qaoa = QAOA()
        qaoa.set_device(device)
        qaoa.set_circuit_properties(
            p=p, param_type="standard", init_type="rand", mixer_hamiltonian="x"
        )
        qaoa.set_backend_properties(prepend_state=None, append_state=None)
        qaoa.set_classical_optimizer(
            method="nelder-mead",
            maxiter=1,
            cost_progress=True,
            parameter_log=True,
            optimization_progress=True,
        )
        qaoa.compile(qubo, routing_function=self.__routing_function_mock)
        qaoa.optimize()

        backend, result = qaoa.backend, qaoa.result

        # do the checks

        assert backend.n_qubits == len(
            expected.final_logical_qubit_order
        ), """Number of qubits in the circuit is not equal to the number of qubits given by routing"""

        assert (
            backend.problem_qubits == qubo.n
        ), f"""Number of nodes in problem is not equal to backend.problem_qubits, 
        is '{backend.problem_qubits }' but should be '{qubo.n}'"""

        assert (
            len(list(result.optimized["measurement_outcomes"].keys())[0]) == qubo.n
        ), "The number of qubits in the optimized circuit is not equal to the number of qubits in the problem."

        # check that swap gates are applied in the correct position
        swap_mask_new = []
        for gate in backend.abstract_circuit:
            if not gate.gate_label.n_qubits == 1:
                swap_mask_new.append(isinstance(gate, SWAPGateMap))

        # create the expected swap mask (for p==2 the second swap mask is reversed)
        expected_swap_mask = []
        for i in range(p):
            expected_swap_mask += expected.swap_mask[:: (-1) ** (i % 2)]

        assert (
            swap_mask_new == expected_swap_mask
        ), "Swap gates are not in the correct position"

        # check that the correct qubits are used in the gates
        gate_indices_list_new = []
        for gate in backend.abstract_circuit:
            if not gate.gate_label.n_qubits == 1:
                gate_indices_list_new.append([gate.qubit_1, gate.qubit_2])

        # create the expected swap mask (for p==2 the second swap mask is reversed)
        expected_gate_indices_list = []
        for i in range(p):
            expected_gate_indices_list += expected.gate_indices_list[:: (-1) ** (i % 2)]

        assert (
            gate_indices_list_new == expected_gate_indices_list
        ), "The qubits used in the gates are not correct"

    @pytest.mark.qpu
    def test_qubit_routing(self):
        for i, case in enumerate(self.list_of_cases):
            print("Test case {} out of {}:".format(i + 1, len(self.list_of_cases)))
            self.__compare_results(case, p=i % 4 + 1)
            print("Test passed for case: {}".format(case.values_input()))


if __name__ == "__main__":
    unittest.main()



import unittest
import numpy as np
import pytest

from pyquil import Program, quilbase
from pyquil.gates import RX, RY, RZ

from openqaoa import QAOA
from openqaoa.qaoa_components import (
    create_qaoa_variational_params,
    QAOADescriptor,
    PauliOp,
    Hamiltonian,
)
from openqaoa.utilities import X_mixer_hamiltonian
from openqaoa.backends import QAOAvectorizedBackendSimulator, create_device
from openqaoa.problems import NumberPartition, QUBO
from openqaoa_pyquil.backends import DevicePyquil, QAOAPyQuilQPUBackend


class TestingQAOACostPyquilQVM(unittest.TestCase):

    """This Object tests the QAOA Cost PyQuil QPU object, which is tasked with the
    creation and execution of a QAOA circuit for the selected QPU provider and
    backend. `as_qvm` is set to be True throughout.

    For all of these tests, qvm and quilc must be running.
    """

    @pytest.mark.qvm
    def test_connection(self):
        """
        Checks if connection to qvm and quilc is successful.
        TODO : improve test
        """

        # Check connection to qvm
        device_pyquil = DevicePyquil(
            device_name="2q-qvm", as_qvm=True, execution_timeout=3, compiler_timeout=3
        )

        # Check connection to quilc compiler
        program = Program().inst(RX(np.pi, 0))
        device_pyquil.quantum_computer.compiler.quil_to_native_quil(program)

        pass

    @pytest.mark.qvm
    def test_active_reset(self):
        """
        Test if active_reset works fine.
        Check for RESET instruction in parametric circuit when active_reset = True / False
        """

        device_pyquil = DevicePyquil(
            device_name="2q-qvm", as_qvm=True, execution_timeout=3, compiler_timeout=3
        )
        cost_hamil = Hamiltonian([PauliOp("Z", (0,)), PauliOp("Z", (1,))], [1, 2], 1)
        mixer_hamil = X_mixer_hamiltonian(n_qubits=2)
        qaoa_descriptor = QAOADescriptor(cost_hamil, mixer_hamil, p=1)

        backend_obj_pyquil = QAOAPyQuilQPUBackend(
            qaoa_descriptor=qaoa_descriptor,
            device=device_pyquil,
            prepend_state=None,
            append_state=None,
            init_hadamard=True,
            cvar_alpha=1,
            n_shots=1,
            active_reset=True,
        )
        assert "RESET" in [
            str(instr) for instr in backend_obj_pyquil.parametric_circuit
        ]

        backend_obj_pyquil = QAOAPyQuilQPUBackend(
            qaoa_descriptor=qaoa_descriptor,
            device=device_pyquil,
            prepend_state=None,
            append_state=None,
            init_hadamard=True,
            cvar_alpha=1,
            n_shots=1,
            active_reset=False,
        )
        assert "RESET" not in [
            str(instr) for instr in backend_obj_pyquil.parametric_circuit
        ]

    @pytest.mark.qvm
    def test_rewiring(self):
        """
        Test if rewiring works fine.

        """

        device_pyquil = DevicePyquil(
            device_name="2q-qvm", as_qvm=True, execution_timeout=3, compiler_timeout=3
        )
        cost_hamil = Hamiltonian([PauliOp("Z", (0,)), PauliOp("Z", (1,))], [1, 2], 1)
        mixer_hamil = X_mixer_hamiltonian(n_qubits=2)
        qaoa_descriptor = QAOADescriptor(cost_hamil, mixer_hamil, p=1)

        # Test if error is raised correctly
        self.assertRaises(
            ValueError,
            lambda: QAOAPyQuilQPUBackend(
                qaoa_descriptor=qaoa_descriptor,
                device=device_pyquil,
                prepend_state=None,
                append_state=None,
                init_hadamard=True,
                cvar_alpha=1,
                n_shots=1,
                rewiring="illegal string",
            ),
        )

        # Test when rewiring = 'PRAGMA INITIAL_REWIRING "NAIVE"'
        backend_obj_pyquil = QAOAPyQuilQPUBackend(
            qaoa_descriptor=qaoa_descriptor,
            device=device_pyquil,
            prepend_state=None,
            append_state=None,
            init_hadamard=True,
            cvar_alpha=1,
            n_shots=1,
            rewiring='PRAGMA INITIAL_REWIRING "NAIVE"',
        )
        assert 'PRAGMA INITIAL_REWIRING "NAIVE"' in [
            str(instr) for instr in backend_obj_pyquil.parametric_circuit
        ]

        # Test when rewiring = 'PRAGMA INITIAL_REWIRING "PARTIAL"'
        backend_obj_pyquil = QAOAPyQuilQPUBackend(
            qaoa_descriptor=qaoa_descriptor,
            device=device_pyquil,
            prepend_state=None,
            append_state=None,
            init_hadamard=True,
            cvar_alpha=1,
            n_shots=1,
            rewiring='PRAGMA INITIAL_REWIRING "PARTIAL"',
        )
        assert 'PRAGMA INITIAL_REWIRING "PARTIAL"' in [
            str(instr) for instr in backend_obj_pyquil.parametric_circuit
        ]

    @pytest.mark.qvm
    def test_qaoa_pyquil_expectation(self):
        """
        Checks if expectation value agrees with known values. Since angles are selected such that the final state is one of the computational basis states, shots do not matter (there is no statistical variance).
        """

        device_pyquil = DevicePyquil(
            device_name="2q-qvm", as_qvm=True, execution_timeout=3, compiler_timeout=3
        )

        # Without interaction terms
        cost_hamil = Hamiltonian([PauliOp("Z", (0,)), PauliOp("Z", (1,))], [1, 1], 1)
        mixer_hamil = X_mixer_hamiltonian(n_qubits=2)
        qaoa_descriptor = QAOADescriptor(cost_hamil, mixer_hamil, p=1)
        variate_params = create_qaoa_variational_params(
            qaoa_descriptor, "standard", "ramp"
        )

        args = [np.pi / 4, np.pi / 4]  # beta, gamma
        variate_params.update_from_raw(args)

        backend_obj_pyquil = QAOAPyQuilQPUBackend(
            qaoa_descriptor=qaoa_descriptor,
            device=device_pyquil,
            prepend_state=None,
            append_state=None,
            init_hadamard=True,
            cvar_alpha=1,
            n_shots=1,
        )
        backend_obj_pyquil.expectation(variate_params)

        assert np.isclose(backend_obj_pyquil.expectation(variate_params), -1)

    @pytest.mark.qvm
    def test_qaoa_pyquil_gate_names(self):
        """
        Checks if names of gates are correct, and no. of measurement gates match the no. of qubits.
        """

        device_pyquil = DevicePyquil(
            device_name="2q-qvm", as_qvm=True, execution_timeout=3, compiler_timeout=3
        )

        # Without interaction terms
        cost_hamil = Hamiltonian([PauliOp("Z", (0,)), PauliOp("Z", (1,))], [1, 1], 1)
        mixer_hamil = X_mixer_hamiltonian(n_qubits=2)
        qaoa_descriptor = QAOADescriptor(cost_hamil, mixer_hamil, p=1)
        params = create_qaoa_variational_params(qaoa_descriptor, "standard", "ramp")
        backend_obj_pyquil = QAOAPyQuilQPUBackend(
            qaoa_descriptor=qaoa_descriptor,
            device=device_pyquil,
            prepend_state=None,
            append_state=None,
            init_hadamard=True,
            cvar_alpha=1,
            n_shots=1,
        )

        gate_names = [
            instr.name
            for instr in backend_obj_pyquil.parametric_circuit
            if type(instr) == quilbase.Gate
        ]
        assert gate_names == [
            "RZ",
            "RX",
            "RZ",
            "RX",
            "RZ",
            "RX",
            "RZ",
            "RX",
            "RZ",
            "RZ",
            "RX",
            "RX",
        ]

        measurement_gate_no = len(
            [
                instr
                for instr in backend_obj_pyquil.qaoa_circuit(params)
                if type(instr) == quilbase.Measurement
            ]
        )
        assert measurement_gate_no == 2

        # With interaction terms
        cost_hamil = Hamiltonian(
            [PauliOp("Z", (0,)), PauliOp("Z", (1,)), PauliOp("ZZ", (0, 1))],
            [1, 1, 1],
            1,
        )
        mixer_hamil = X_mixer_hamiltonian(n_qubits=2)
        qaoa_descriptor = QAOADescriptor(cost_hamil, mixer_hamil, p=1)
        params = create_qaoa_variational_params(qaoa_descriptor, "standard", "ramp")
        backend_obj_pyquil = QAOAPyQuilQPUBackend(
            qaoa_descriptor=qaoa_descriptor,
            device=device_pyquil,
            prepend_state=None,
            append_state=None,
            init_hadamard=True,
            cvar_alpha=1,
            n_shots=1,
        )

        gate_names = [
            instr.name
            for instr in backend_obj_pyquil.parametric_circuit
            if type(instr) == quilbase.Gate
        ]
        assert gate_names == [
            "RZ",
            "RX",
            "RZ",
            "RX",
            "RZ",
            "RX",
            "RZ",
            "RX",
            "RZ",
            "RZ",
            "RZ",
            "RZ",
            "CPHASE",
            "RX",
            "RX",
        ]

        measurement_gate_no = len(
            [
                instr
                for instr in backend_obj_pyquil.qaoa_circuit(params)
                if type(instr) == quilbase.Measurement
            ]
        )
        assert measurement_gate_no == 2

    @pytest.mark.qvm
    def test_circuit_init_hadamard(self):
        """
        Checks correctness of circuit for the argument `init_hadamard`.
        """

        device_pyquil = DevicePyquil(
            device_name="2q-qvm", as_qvm=True, execution_timeout=3, compiler_timeout=3
        )

        # With hadamard
        cost_hamil = Hamiltonian(
            [PauliOp("Z", (0,)), PauliOp("Z", (1,)), PauliOp("ZZ", (0, 1))],
            [1, 1, 1],
            1,
        )
        mixer_hamil = X_mixer_hamiltonian(n_qubits=2)
        qaoa_descriptor = QAOADescriptor(cost_hamil, mixer_hamil, p=1)
        pyquil_backend = QAOAPyQuilQPUBackend(
            device_pyquil,
            qaoa_descriptor,
            n_shots=10,
            prepend_state=None,
            append_state=None,
            init_hadamard=True,
            cvar_alpha=1,
        )

        assert [
            "RZ",
            "RX",
            "RZ",
            "RX",
            "RZ",
            "RX",
            "RZ",
            "RX",
            "RZ",
            "RZ",
            "RZ",
            "RZ",
            "CPHASE",
            "RX",
            "RX",
        ] == [
            instr.name
            for instr in pyquil_backend.parametric_circuit
            if type(instr) == quilbase.Gate
        ]

        # Without hadamard
        cost_hamil = Hamiltonian(
            [PauliOp("Z", (0,)), PauliOp("Z", (1,)), PauliOp("ZZ", (0, 1))],
            [1, 1, 1],
            1,
        )
        mixer_hamil = X_mixer_hamiltonian(n_qubits=2)
        qaoa_descriptor = QAOADescriptor(cost_hamil, mixer_hamil, p=1)
        pyquil_backend = QAOAPyQuilQPUBackend(
            device_pyquil,
            qaoa_descriptor,
            n_shots=10,
            prepend_state=None,
            append_state=None,
            init_hadamard=False,
            cvar_alpha=1,
        )

        assert ["RZ", "RZ", "RZ", "RZ", "CPHASE", "RX", "RX"] == [
            instr.name
            for instr in pyquil_backend.parametric_circuit
            if type(instr) == quilbase.Gate
        ]

    @pytest.mark.qvm
    def test_circuit_append_state(self):
        """
        Checks correctness of circuit for the argument `append_state`.
        """

        device_pyquil = DevicePyquil(
            device_name="2q-qvm", as_qvm=True, execution_timeout=3, compiler_timeout=3
        )

        # With append_state
        cost_hamil = Hamiltonian(
            [PauliOp("Z", (0,)), PauliOp("Z", (1,)), PauliOp("ZZ", (0, 1))],
            [1, 1, 1],
            1,
        )
        mixer_hamil = X_mixer_hamiltonian(n_qubits=2)
        qaoa_descriptor = QAOADescriptor(cost_hamil, mixer_hamil, p=1)

        params = create_qaoa_variational_params(qaoa_descriptor, "standard", "ramp")

        append_circuit = Program().inst(RX(np.pi, 0), RY(np.pi / 2, 1), RZ(np.pi, 0))

        pyquil_backend = QAOAPyQuilQPUBackend(
            device_pyquil,
            qaoa_descriptor,
            n_shots=10,
            prepend_state=None,
            append_state=append_circuit,
            init_hadamard=False,
            cvar_alpha=1,
        )

        assert set(
            [
                "RZ",
                "RZ",
                "RX",
                "RZ",
                "RX",
                "CZ",
                "RZ",
                "RZ",
                "RX",
                "RZ",
                "RX",
                "RZ",
                "RX",
                "RZ",
                "RX",
                "CZ",
                "RZ",
                "RX",
                "RZ",
                "RX",
                "RZ",
                "RX",
                "RZ",
                "RZ",
                "RX",
                "RZ",
                "RX",
                "RZ",
                "RZ",
                "RX",
                "RZ",
                "RZ",
                "RX",
                "RZ",
                "RX",
                "RZ",
            ]
        ) == set(
            [
                instr.name
                for instr in pyquil_backend.qaoa_circuit(params)
                if type(instr) == quilbase.Gate
            ]
        )

    @pytest.mark.qvm
    def test_circuit_prepend_state(self):
        """
        Checks correctness of circuit for the argument `prepend_state`.
        """

        device_pyquil = DevicePyquil(
            device_name="2q-qvm", as_qvm=True, execution_timeout=3, compiler_timeout=3
        )

        # With prepend_state
        cost_hamil = Hamiltonian(
            [PauliOp("Z", (0,)), PauliOp("Z", (1,)), PauliOp("ZZ", (0, 1))],
            [1, 1, 1],
            1,
        )
        mixer_hamil = X_mixer_hamiltonian(n_qubits=2)
        qaoa_descriptor = QAOADescriptor(cost_hamil, mixer_hamil, p=1)

        prepend_circuit = Program().inst(RX(np.pi, 0), RY(np.pi / 2, 1), RZ(np.pi, 0))

        pyquil_backend = QAOAPyQuilQPUBackend(
            device_pyquil,
            qaoa_descriptor,
            n_shots=10,
            prepend_state=prepend_circuit,
            append_state=None,
            init_hadamard=False,
            cvar_alpha=1,
        )

        assert ["RX", "RY", "RZ", "RZ", "RZ", "RZ", "RZ", "CPHASE", "RX", "RX"] == [
            instr.name
            for instr in pyquil_backend.parametric_circuit
            if type(instr) == quilbase.Gate
        ]

        # Test if error is raised correctly
        prepend_circuit = Program().inst(RX(np.pi, 0), RY(np.pi / 2, 1), RZ(np.pi, 2))
        self.assertRaises(
            AssertionError,
            lambda: QAOAPyQuilQPUBackend(
                qaoa_descriptor=qaoa_descriptor,
                device=device_pyquil,
                prepend_state=prepend_circuit,
                append_state=None,
                init_hadamard=True,
                cvar_alpha=1,
                n_shots=1,
            ),
        )

    @pytest.mark.qvm
    def test_pyquil_vectorized_agreement(self):
        """
        Checks correctness of expectation values with vectorized backend, up to a tolerance of delta = std.dev.
        """

        # Without interaction terms
        device_pyquil = DevicePyquil(
            device_name="2q-qvm", as_qvm=True, execution_timeout=3, compiler_timeout=3
        )
        device_pyquil.quantum_computer.qam.random_seed = 1

        cost_hamil = Hamiltonian(
            [PauliOp("Z", (0,)), PauliOp("Z", (1,)), PauliOp("ZZ", (0, 1))],
            [1, 1, 1],
            1,
        )
        mixer_hamil = X_mixer_hamiltonian(n_qubits=2)
        qaoa_descriptor = QAOADescriptor(cost_hamil, mixer_hamil, p=1)
        variate_params = create_qaoa_variational_params(
            qaoa_descriptor, "standard", "ramp"
        )
        args = [np.pi / 8, np.pi / 4]  # beta, gamma

        variate_params.update_from_raw(args)
        backend_obj_pyquil = QAOAPyQuilQPUBackend(
            qaoa_descriptor=qaoa_descriptor,
            device=device_pyquil,
            prepend_state=None,
            append_state=None,
            init_hadamard=True,
            cvar_alpha=1,
            n_shots=10,
        )
        expt_pyquil = backend_obj_pyquil.expectation(variate_params)

        variate_params.update_from_raw(args)
        backend_obj_vectorized = QAOAvectorizedBackendSimulator(
            qaoa_descriptor, prepend_state=None, append_state=None, init_hadamard=True
        )
        expt_vec, std_dev_vec = backend_obj_vectorized.expectation_w_uncertainty(
            variate_params
        )

        self.assertAlmostEqual(expt_vec, expt_pyquil, delta=std_dev_vec)

    @pytest.mark.qvm
    def test_remote_qubit_overflow(self):
        """
        If the user creates a circuit that is larger than the maximum circuit size
        that is supported by the QPU. An Exception should be raised with the
        appropriate error message alerting the user to the error.
        """

        shots = 100
        set_of_numbers = np.random.randint(1, 10, 6).tolist()
        qubo = NumberPartition(set_of_numbers).qubo
        mixer_hamil = X_mixer_hamiltonian(n_qubits=6)
        qaoa_descriptor = QAOADescriptor(qubo.hamiltonian, mixer_hamil, p=1)
        variate_params = create_qaoa_variational_params(
            qaoa_descriptor, "standard", "rand"
        )

        device_qvm = DevicePyquil(
            device_name="5q-qvm", as_qvm=True, execution_timeout=3, compiler_timeout=3
        )
        try:
            qvm_backend = QAOAPyQuilQPUBackend(
                device_qvm, qaoa_descriptor, shots, None, None, True, 1.0
            )
            qvm_backend.expectation(variate_params)
        except Exception as e:
            self.assertEqual(
                str(e),
                "There are lesser qubits on the device than the number of qubits required for the circuit.",
            )

    @pytest.mark.qvm
    def test_job_ids(self):
        """
        Test if correct job ids are generated and returned when running on qvm
        """

        # define problem
        problem = QUBO.random_instance(3)

        # initialize
        q = QAOA()

        # device
        device = create_device(location="qcs", name="3q-qvm")
        q.set_device(device)

        # classical optimizer only 3 iterations
        q.set_classical_optimizer(maxiter=3)

        # compile
        q.compile(problem)

        # run
        q.optimize()

        # check if we have job ids
        opt_id = q.result.optimized["job_id"]
        assert len(opt_id) == 36 and isinstance(
            opt_id, str
        ), f"QCS QVM: job id is not a string of length 36, but {opt_id}"

        inter_id = q.result.intermediate["job_id"]
        for id in inter_id:
            assert len(id) == 36 and isinstance(
                id, str
            ), f"QCS QVM: on intermediate job id is not a string of length 36, but {id}"


if __name__ == "__main__":
    unittest.main()



from setuptools import setup, find_packages
from os import getcwd

current_path = getcwd()

with open("README.md", "r") as fh:
    long_description = fh.read()

with open("openqaoa_pyquil/_version.py") as f:
    version = f.readlines()[-1].split()[-1].strip("\"'")

requirements = [
    "openqaoa-core=={}".format(version),
]

add_requirements = open("requirements.txt").readlines()
add_requirements = [r.strip() for r in add_requirements]

requirements.extend(add_requirements)

setup(
    name="openqaoa-pyquil",
    python_requires=">=3.8, <3.11",
    version=version,
    author="Entropica Labs",
    packages=find_packages(where="."),
    entry_points={"openqaoa.plugins": ["pyquil = openqaoa_pyquil.backend_config"]},
    url="https://github.com/entropicalabs/openqaoa",
    install_requires=requirements,
    license="MIT",
    description="Pyquil Plug-in for OpenQAOA",
    long_description=long_description,
    long_description_content_type="text/markdown",
    classifiers=[
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
        "Programming Language :: Python :: 3.10",
        "Operating System :: OS Independent",
    ],
    keywords="quantum optimisation SDK",
)



from collections import Counter
import numpy as np
from copy import deepcopy
from pyquil import Program, gates, quilbase
from typing import List, Optional
import warnings

from .devices import DevicePyquil
from .gates_pyquil import PyquilGateApplicator
from openqaoa.backends.basebackend import (
    QAOABaseBackendShotBased,
    QAOABaseBackendCloud,
    QAOABaseBackendParametric,
)
from openqaoa.qaoa_components import QAOADescriptor
from openqaoa.qaoa_components.variational_parameters.variational_baseparams import (
    QAOAVariationalBaseParams,
)
from openqaoa.qaoa_components.ansatz_constructor.gatemap import RZZGateMap, SWAPGateMap
from openqaoa.qaoa_components.ansatz_constructor.rotationangle import RotationAngle
from openqaoa.utilities import generate_uuid


def check_edge_connectivity(executable: Program, device: DevicePyquil):
    '''
    Check that the program does not contain 2-qubit terms that is not present
    in the QPU's topology (to prevent quilc from crashing).

    Parameters
    ----------
    executable: `Program`
        pyQuil executable program.
    device: `DevicePyquil`
        An object of the class ``DevicePyquil`` which contains information on
        pyQuil's `QuantumComputer` object, used to extract the selected QPU's topology.

    Returns
    -------
        None
    """
    '''

    qpu_graph = device.quantum_computer.qubit_topology()

    instrs = [instr for instr in executable if type(instr) == quilbase.Gate]
    pair_instrs = [
        list(instr.get_qubits()) for instr in instrs if len(instr.get_qubits()) == 2
    ]

    for term in pair_instrs:
        if len(term) == 2:
            assert (
                term in qpu_graph.edges()
            ), f"Term {term} is not an edge on the QPU graph of {device.device_name}."


class QAOAPyQuilQPUBackend(
    QAOABaseBackendParametric, QAOABaseBackendCloud, QAOABaseBackendShotBased
):
    """
    A QAOA backend object for real Rigetti QPUs

    Parameters
    ----------
    device: `DevicePyquil`
        The device object to access pyquil devices with credentials.
    qaoa_descriptor: `QAOADescriptor`
        An object of the class ``QAOADescriptor`` which contains information on
        circuit construction and depth of the circuit.
    n_shots: `int`
        The number of shots to be taken for each circuit.
    prepend_state: `pyquil.Program`
        The state prepended to the circuit.
    append_state: `pyquil.Program`
        The state appended to the circuit.
    init_hadamard: `bool`
        Whether to apply a Hadamard gate to the beginning of the QAOA part of the circuit.
    cvar_alpha: `float`
        Conditional Value-at-Risk (CVaR) – a measure that takes into account only the tail of the
        probability distribution arising from the circut's count dictionary. Must be between 0 and 1. Check
        https://arxiv.org/abs/1907.04769 for further details.
    active_reset:
        Whether to use the pyQuil's active reset scheme to reset qubits between shots.
    rewiring:
        Rewiring scheme to be used for Pyquil.
        Either PRAGMA INITIAL_REWIRING "NAIVE" or PRAGMA INITIAL_REWIRING "PARTIAL".
        If None, pyquil defaults according to:
        NAIVE: The qubits used in all instructions in the program satisfy the
        topological constraints of the device.
        PARTIAL: Otherwise.
    """

    def __init__(
        self,
        device: DevicePyquil,
        qaoa_descriptor: QAOADescriptor,
        n_shots: int,
        prepend_state: Program,
        append_state: Program,
        init_hadamard: bool,
        cvar_alpha: float,
        active_reset: bool = False,
        rewiring: str = "",
        initial_qubit_mapping: Optional[List[int]] = None,
    ):
        QAOABaseBackendShotBased.__init__(
            self,
            qaoa_descriptor,
            n_shots,
            prepend_state,
            append_state,
            init_hadamard,
            cvar_alpha,
        )
        QAOABaseBackendCloud.__init__(self, device)

        self.gate_applicator = PyquilGateApplicator()

        self.active_reset = active_reset
        self.rewiring = rewiring
        self.qureg = list(range(self.n_qubits))
        self.problem_reg = self.qureg[0 : self.problem_qubits]

        if self.initial_qubit_mapping is None:
            self.initial_qubit_mapping = (
                initial_qubit_mapping
                if initial_qubit_mapping is not None
                else list(range(self.n_qubits))
            )
        else:
            if isinstance(initial_qubit_mapping, list):
                warnings.warn(
                    "Ignoring the initial_qubit_mapping since the routing algorithm chose one"
                )

        # self.qureg_placeholders = QubitPlaceholder.register(self.n_qubits)
        self.qubit_mapping = dict(zip(self.qureg, self.initial_qubit_mapping))

        if self.prepend_state:
            assert self.n_qubits >= len(prepend_state.get_qubits()), (
                "Cannot attach a bigger circuit " "to the QAOA routine"
            )

        if self.device.n_qubits < self.n_qubits:
            raise Exception(
                "There are lesser qubits on the device than the number of qubits required for the circuit."
            )

        self.parametric_circuit = self.parametric_qaoa_circuit

        # Check program connectivity against QPU connectivity
        # TODO: reconcile with PRAGMA PRESERVE

        # check_edge_connectivity(self.prog_exe, device)

    def qaoa_circuit(self, params: QAOAVariationalBaseParams) -> Program:
        """
        Injects angles into created executable parametric circuit.

        Parameters
        ----------
        params: `QAOAVariationalBaseParams`

        Returns
        -------
        `pyquil.Program`
            A pyquil.Program (executable) object.
        """
        parametric_circuit = deepcopy(self.parametric_circuit)
        # declare the read-out register
        ro = parametric_circuit.declare("ro", "BIT", self.problem_qubits)

        if self.append_state:
            parametric_circuit += self.append_state

        if self.final_mapping is None:
            for i, qbit in enumerate(self.problem_reg):
                parametric_circuit += gates.MEASURE(self.qubit_mapping[qbit], ro[i])
        else:
            # Measurement instructions
            for i, qubit in enumerate(self.final_mapping[0 : len(self.problem_reg)]):
                cbit = ro[i]
                parametric_circuit += gates.MEASURE(self.qubit_mapping[qubit], cbit)
        parametric_circuit.wrap_in_numshots_loop(self.n_shots)

        native = self.device.quantum_computer.compiler.quil_to_native_quil(
            parametric_circuit
        )

        prog_exe = self.device.quantum_computer.compiler.native_quil_to_executable(
            native
        )

        angles_list = np.array(
            self.obtain_angles_for_pauli_list(self.abstract_circuit, params),
            dtype=float,
        )
        angle_declarations = list(parametric_circuit.declarations.keys())

        angle_declarations.remove("ro")

        for i, param_name in enumerate(angle_declarations):
            prog_exe.write_memory(region_name=param_name, value=angles_list[i])

        return prog_exe

    @property
    def parametric_qaoa_circuit(self) -> Program:
        """
        Creates a parametric QAOA circuit (pyquil.Program object),
        given the qubit pairs, single qubits with biases,
        and a set of circuit angles. Note that this function does not actually run
        the circuit.

        Parameters
        ----------
        params: `QAOAVariationalBaseParams`

        Returns
        -------
        `pyquil.Program`
            A pyquil.Program object.
        """

        if self.active_reset:
            parametric_circuit = Program(gates.RESET())
        else:
            parametric_circuit = Program()

        if self.rewiring != None:
            if self.rewiring in [
                'PRAGMA INITIAL_REWIRING "NAIVE"',
                'PRAGMA INITIAL_REWIRING "PARTIAL"',
                "",
            ]:
                parametric_circuit += Program(self.rewiring)
            else:
                raise ValueError(
                    "Rewiring command not recognized. Please use "
                    'PRAGMA INITIAL_REWIRING "NAIVE"'
                    " or "
                    'PRAGMA INITIAL_REWIRING "PARTIAL"'
                    ""
                )

        if self.prepend_state:
            parametric_circuit += self.prepend_state

        # Initial state is all |+>
        if self.init_hadamard:
            for i in self.problem_reg:
                parametric_circuit += gates.RZ(np.pi, self.qubit_mapping[i])
                parametric_circuit += gates.RX(np.pi / 2, self.qubit_mapping[i])
                parametric_circuit += gates.RZ(np.pi / 2, self.qubit_mapping[i])
                parametric_circuit += gates.RX(-np.pi / 2, self.qubit_mapping[i])

        # create a list of gates in order of application on quantum circuit
        for each_gate in self.abstract_circuit:
            # if gate is of type mixer or cost gate, assign parameter to it
            if each_gate.gate_label.type.value in ["MIXER", "COST"]:
                gatelabel_pyquil = each_gate.gate_label.__repr__()
                gatelabel_pyquil = (
                    "one" + gatelabel_pyquil[1:]
                    if each_gate.gate_label.n_qubits == 1
                    else "two" + gatelabel_pyquil[1:]
                )
                angle_param = parametric_circuit.declare(gatelabel_pyquil, "REAL", 1)
                each_gate.angle_value = angle_param
            if isinstance(each_gate, RZZGateMap) or isinstance(each_gate, SWAPGateMap):
                decomposition = each_gate.decomposition("standard2")
            else:
                decomposition = each_gate.decomposition("standard")

            # using the list above, construct the circuit
            for each_tuple in decomposition:
                if type(each_tuple[1][-1]) == RotationAngle:
                    rotation_angle = each_tuple[1][-1]
                    qubits = each_tuple[1][:-1]
                else:
                    rotation_angle = None
                    qubits = each_tuple[1]
                if not isinstance(qubits, list):
                    qubits = [qubits]
                new_qubits = [self.qubit_mapping[qubit] for qubit in qubits]

                if rotation_angle is None:
                    gate = each_tuple[0](self.gate_applicator, *new_qubits)
                else:
                    gate = each_tuple[0](
                        self.gate_applicator, *new_qubits, rotation_angle
                    )
                gate.apply_gate(parametric_circuit)

        return parametric_circuit

    def get_counts(self, params: QAOAVariationalBaseParams, n_shots=None) -> dict:
        """
        Execute the circuit and obtain the counts.

        Parameters
        ----------
        params: QAOAVariationalBaseParams
            The QAOA parameters - an object of one of the parameter classes, containing
            variable parameters.
        n_shots: int
            The number of times to run the circuit. If None, n_shots is set to the default: self.n_shots

        Returns
        -------
        counts : dictionary
            A dictionary with the bitstring as the key and the number of counts as its value.
        """

        executable_program = self.qaoa_circuit(params)

        # if n_shots is given change the number of shots
        if n_shots is not None:
            executable_program.wrap_in_numshots_loop(n_shots)

        result = self.device.quantum_computer.run(executable_program)
        # we create an uuid for the job
        self.job_id = generate_uuid()

        # TODO: check the endian (big or little) ordering of measurement outcomes
        meas_list = [
            "".join(str(bit) for bit in bitstring)
            for bitstring in result.readout_data["ro"]
        ]

        # Expose counts
        final_counts = Counter(list(meas_list))
        self.measurement_outcomes = final_counts
        return final_counts

    def circuit_to_qasm(self, params: QAOAVariationalBaseParams) -> str:
        """
        A method to convert the pyQuil program to a OpenQASM string.
        """
        raise NotImplementedError()
        # qasm_program = self.device.quantum_computer.compiler.quil_to_qasm(self.qaoa_circuit(params))
        # return qasm_program

    def reset_circuit(self):
        """
        Reset self.program after performing a computation. Also handle active reset and rewirings.
        """
        pass
